<?php
/**
 * Order admin handler.
 *
 * @package WooCommerce_Gateway_GoCardless
 */

/**
 * Manage order admin UI for orders paid via GoCardless.
 *
 * @since 2.4.0
 */
class WC_GoCardless_Order_Admin {

	/**
	 * Register webhook events meta box.
	 *
	 * @since 2.4.0
	 */
	public function add_meta_box() {
		add_action( 'add_meta_boxes', array( $this, 'add_webhook_events_meta_box' ), 11, 2 );
	}

	/**
	 * Add the webhook events meta box.
	 *
	 * @since 2.4.0
	 *
	 * @param string           $post_type Post Type.
	 * @param WP_Post|WC_Order $post      Order or Post object.
	 */
	public function add_webhook_events_meta_box( $post_type, $post ) {
		$order = ( $post instanceof WP_Post ) ? wc_get_order( $post->ID ) : $post;
		if ( ! $order ) {
			return false;
		}

		if ( 'gocardless' !== wc_gocardless_get_order_prop( $order, 'payment_method' ) ) {
			return false;
		}

		$screen = WC_GoCardless_Compat::is_cot_enabled() ? wc_get_page_screen_id( 'shop-order' ) : 'shop_order';

		add_meta_box( 'woocommerce-gocardless-webhook-events', esc_html__( 'GoCardless Webhook Events', 'woocommerce-gateway-gocardless' ), array( $this, 'webhook_events_meta_box' ), $screen, 'side' );
	}

	/**
	 * Output the content for webhook events meta box.
	 *
	 * @since 2.4.0
	 *
	 * @param WP_Post|WC_Order $post Order or Post object.
	 */
	public function webhook_events_meta_box( $post ) {
		$order = ( $post instanceof WP_Post ) ? wc_get_order( $post->ID ) : $post;
		if ( ! $order ) {
			return false;
		}

		$events = $order->get_meta( '_gocardless_webhook_events', false );

		// TODO(gedex): This will makes the list long for subscription payments,
		// paginate should be implemented.
		//phpcs:disable WordPress.Security.EscapeOutput.OutputNotEscaped
		if ( ! empty( $events ) ) {
			echo wpautop( esc_html__( 'This list may contains duplicated events.', 'woocommerce-gateway-gocardless' ) );
			echo '<ul class="order_notes">';
			foreach ( $events as $event_obj ) {
				$event_data = $event_obj->get_data();
				$event      = $event_data['value'];
				printf( '<li rel="%s">%s</li>', $event['id'], $this->get_formatted_event_item( $event ) );
			}
			echo '</ul>';
		} else {
			echo wpautop( esc_html__( 'No recent events.', 'woocommerce-gateway-gocardless' ) );
		}
		//phpcs:enable WordPress.Security.EscapeOutput.OutputNotEscaped
	}

	/**
	 * Get formatted event item.
	 *
	 * @since 2.4.0
	 *
	 * @param array $event Event array.
	 *
	 * @return string Formatted event item.
	 */
	public function get_formatted_event_item( $event ) {
		$resource_id = '';
		if ( ! empty( $event['links']['mandate'] ) ) {
			$resource_id = $event['links']['mandate'];
		} elseif ( ! empty( $event['links']['payment'] ) ) {
			$resource_id = $event['links']['payment'];
		} elseif ( ! empty( $event['links']['subscription'] ) ) {
			$resource_id = $event['links']['subscription'];
		} elseif ( ! empty( $event['links']['refund'] ) ) {
			$resource_id = $event['links']['refund'];
		}

		return sprintf(
			'
			<div class="note_content"><p><strong>%s%s</strong></p>%s</div>
			<p class="meta">
				<abbr class="exact-date">%s</abbr>
			</p>
			',
			$event['resource_type'] . ' ' . $event['action'],
			! empty( $resource_id ) ? ' → ' . $resource_id : '',
			! empty( $event['details']['description'] ) ? wpautop( $event['details']['description'] ) : '',
			/* translators: Logged on time */
			sprintf( esc_html__( 'Logged on %s', 'woocommerce-gateway-gocardless' ), $event['created_at'] )
		);
	}

	/**
	 * Add GoCardless specific order actions.
	 *
	 * @since 2.4.0
	 */
	public function add_order_actions() {
		add_action( 'woocommerce_order_actions', array( $this, 'gocardless_actions' ), 10, 2 );
		add_action( 'woocommerce_order_action_gocardless_cancel_payment', array( $this, 'cancel_payment' ) );
		add_action( 'woocommerce_order_action_gocardless_retry_payment', array( $this, 'retry_payment' ) );
	}

	/**
	 * GoCardless order actions.
	 *
	 * @since 2.4.0
	 *
	 * @param array    $actions The actions available for the order.
	 * @param WC_Order $order   WC_Order object.
	 *
	 * @return array Order actions
	 */
	public function gocardless_actions( $actions, $order ) {
		if ( empty( $order ) ) {
			return $actions;
		}

		if ( 'gocardless' !== wc_gocardless_get_order_prop( $order, 'payment_method' ) ) {
			return $actions;
		}

		$gateway = wc_gocardless()->gateway_instance();
		if ( ! $gateway ) {
			return $actions;
		}

		$order_id = wc_gocardless_get_order_prop( $order, 'id' );

		// For payment with status pending_submission, merchant can cancel it.
		$payment_status = $gateway->get_order_resource( $order_id, 'payment', 'status' );
		switch ( $payment_status ) {
			case 'pending_submission':
				$actions['gocardless_cancel_payment'] = esc_html__( 'GoCardless &rsaquo; Cancel payment', 'woocommerce-gateway-gocardless' );
				// Display action text to "Cancel Payment & Subscription" for subscription and parent order.
				// @see https://github.com/woocommerce/woocommerce-gateway-gocardless/issues/123.
				if (
					function_exists( 'wcs_is_subscription' ) &&
					(
						wcs_is_subscription( $order_id ) ||
						! empty( wcs_get_subscriptions_for_order( $order, array( 'order_type' => 'parent' ) ) )
					)
				) {
					$actions['gocardless_cancel_payment'] = esc_html__( 'GoCardless &rsaquo; Cancel Payment & Subscription', 'woocommerce-gateway-gocardless' );
				}
				break;
			case 'failed':
				$actions['gocardless_retry_payment'] = esc_html__( 'GoCardless &rsaquo; Retry payment', 'woocommerce-gateway-gocardless' );
				break;
		}

		return $actions;
	}

	/**
	 * Order action to cancel payment.
	 *
	 * @since 2.4.0
	 *
	 * @param WC_Order $order WC_Order object.
	 */
	public function cancel_payment( $order ) {
		if ( ! $order ) {
			return;
		}

		if ( 'gocardless' !== wc_gocardless_get_order_prop( $order, 'payment_method' ) ) {
			return;
		}

		$gateway = wc_gocardless()->gateway_instance();
		if ( ! $gateway ) {
			return;
		}

		$order_id   = wc_gocardless_get_order_prop( $order, 'id' );
		$payment_id = $gateway->get_order_resource( $order_id, 'payment', 'id' );
		$payment    = WC_GoCardless_API::cancel_payment( $payment_id );

		if ( class_exists( 'WC_Admin_Notices' ) ) {
			WC_Admin_Notices::add_custom_notice( 'gocardless_cancel_payment', $this->get_cancel_payment_notice( $order->get_order_number(), $payment ) );
		}

		if ( ! is_wp_error( $payment ) ) {
			$order->update_status( 'cancelled', esc_html__( 'Cancelled GoCardless payment.', 'woocommerce-gateway-gocardless' ) );
		}
	}

	/**
	 * Get notice for cancel payment action.
	 *
	 * @since 2.4.0
	 *
	 * @param int            $order_number Order Number.
	 * @param WP_Error|array $payment      Payment array of WP_Error.
	 *
	 * @return string Notice message
	 * TODO:123
	 */
	public function get_cancel_payment_notice( $order_number, $payment ) {
		if ( is_wp_error( $payment ) ) {
			// translators: %1$s: Order Number, %2$s: Error message.
			return sprintf( esc_html__( 'Failed to cancel GoCardless payment in order #%1$s: %2$s', 'woocommerce-gateway-gocardless' ), $order_number, $payment->get_error_message() );
		}

		// translators: %s: Order Number.
		return sprintf( esc_html__( 'GoCardless payment in order #%s is cancelled.', 'woocommerce-gateway-gocardless' ), $order_number );
	}

	/**
	 * Order action to retry payment.
	 *
	 * @since 2.4.0
	 *
	 * @param WC_Order $order WC_Order object.
	 */
	public function retry_payment( $order ) {
		if ( ! $order ) {
			return;
		}

		if ( 'gocardless' !== wc_gocardless_get_order_prop( $order, 'payment_method' ) ) {
			return;
		}

		$gateway = wc_gocardless()->gateway_instance();
		if ( ! $gateway ) {
			return;
		}

		$order_id   = wc_gocardless_get_order_prop( $order, 'id' );
		$payment_id = $gateway->get_order_resource( $order_id, 'payment', 'id' );
		$payment    = WC_GoCardless_API::retry_payment( $payment_id );

		if ( class_exists( 'WC_Admin_Notices' ) ) {
			WC_Admin_Notices::add_custom_notice( 'gocardless_retry_payment', $this->get_retry_payment_notice( $order->get_order_number(), $payment ) );
		}

		if ( ! is_wp_error( $payment ) ) {
			/**
			 * Filter GoCardless Retry Payment order status.
			 *
			 * Allow other plugins to modify order status for GoCardless Retry Payment.
			 *
			 * @param string $status   Order Status.
			 * @param int    $order_id Order ID.
			 */
			$status = apply_filters( 'woocommerce_gocardless_retry_payment_order_status', 'on-hold', $order_id );
			$order->update_status( $status, esc_html__( 'Retried GoCardless payment.', 'woocommerce-gateway-gocardless' ) );
		}
	}

	/**
	 * Get notice for retry payment action.
	 *
	 * @since 2.4.0
	 *
	 * @param int            $order_number Order Number.
	 * @param WP_Error|array $payment      Payment array of WP_Error.
	 *
	 * @return string Notice message
	 */
	public function get_retry_payment_notice( $order_number, $payment ) {
		if ( is_wp_error( $payment ) ) {
			// translators: %1$s: Order Number, %2$s: Error message.
			return sprintf( esc_html__( 'Failed to retry GoCardless payment in order #%1$s: %2$s', 'woocommerce-gateway-gocardless' ), $order_number, $payment->get_error_message() );
		}

		// translators: %s: Order Number.
		return sprintf( esc_html__( 'Retried GoCardless payment in order #%s.', 'woocommerce-gateway-gocardless' ), $order_number );
	}
}
